from .real_ffts import _RealFFTBasis
import numpy as np


class PrimitiveBasis(_RealFFTBasis):

    r"""
    Discretize the microstructure function into `n_states` local states such
    that:

    .. math::

       \frac{1}{\Delta x} \int_{H} \int_{s} \Lambda(h - l)
       m(h, x) dx dh = m[l, s]

    where :math:`\Lambda` is the primitive basis (also called hat basis)
    function

    .. math::

       \Lambda (h - l) = max \Bigg (1-\Bigg |\frac{h(L - 1)}{H} -
       \frac{Hl}{L-1} \Bigg|, 0\Bigg)

    A microstructure function discretized with this basis is subject to the
    following constraint

    .. math::

       \sum_{l=0}^L m[l, s] = 1

    which is equivalent of saying that every location is filled with some
    configuration of local states.

    Here is an example with 3 discrete local states in a microstructure.

    >>> X = np.array([[[1, 1, 0],
    ...                [1, 0 ,2],
    ...                [0, 1, 0]]])
    >>> assert(X.shape == (1, 3, 3))

    The when a microstructure is discretized, the different local states are
    mapped into local state space, which results in an array of shape
    `(n_samples, n_x, n_y, n_states)`, where `n_states=3` in this case.
    For example, if a cell has a label of 2, its local state will be
    `[0, 0, 1]`. The local state can only have values of 0 or 1.

    >>> prim_basis = PrimitiveBasis(n_states=3)
    >>> X_prim = np.array([[[[0, 1, 0],
    ...                      [0, 1, 0],
    ...                      [1, 0, 0]],
    ...                     [[0, 1, 0],
    ...                      [1, 0, 0],
    ...                      [0, 0, 1]],
    ...                     [[1, 0, 0],
    ...                      [0, 1, 0],
    ...                      [1, 0, 0]]]])
    >>> assert(np.allclose(X_prim, prim_basis.discretize(X)))

    Check that the basis works when all the states are present in the
    microstructure.

    >>> prim_basis = PrimitiveBasis(n_states=3)
    >>> X = np.array([1, 1, 0])
    >>> X_prim = np.array([[0, 1, 0],
    ...                    [0, 1, 0],
    ...                    [1, 0, 0]])
    >>> assert(np.allclose(X_prim, prim_basis.discretize(X)))

    In previous two microstructures had values that fell on the peak of the
    primitive (or hat) basis functions. If a local state value falls between
    two peaks of the primitive basis functions the value will be shared by both
    basis functions. To ensure that all local states fall between the peaks
    of two basis functions, we need to specify the local state domain. For
    example, if a cell has a value of 0.4, and the basis has n_states=2 and
    the domain=[0, 1] then the local state is (0.6, 0.4) (the local state
    must sum to 1).

    Here are a few examples where the local states fall between the picks of
    local states. The first specifies the local state space domain between
    `[0, 1]`.

    >>> n_states = 10
    >>> np.random.seed(4)
    >>> X = np.random.random((2, 5, 3, 2))
    >>> X_ = PrimitiveBasis(n_states, [0, 1]).discretize(X)
    >>> H = np.linspace(0, 1, n_states)
    >>> Xtest = np.sum(X_ * H[None,None,None,:], axis=-1)
    >>> assert np.allclose(X, Xtest)

    Here is an example where the local state space domain is between `[-1, 1]`.

    >>> n_states = 3
    >>> X = np.array([-1, 0, 1, 0.5])
    >>> X_test = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0.5, 0.5]]
    >>> X_ = PrimitiveBasis(n_states, [-1, 1]).discretize(X)
    >>> assert np.allclose(X_, X_test)

    If the local state values in the microstructure are outside of the domain
    they can no longer be represented by two primitive basis functions and
    violates constraint above.

    >>> n_states = 2
    >>> X = np.array([-1, 1])
    >>> prim_basis = PrimitiveBasis(n_states, domain=[0, 1])
    >>> prim_basis.discretize(X)
    Traceback (most recent call last):
    ...
    RuntimeError: X must be within the specified domain

    """

    def _select_slice(self, ijk, s0):
        """
        Helper method used to calibrate influence coefficients from in
        mks_localization_model to account for redundancies from linearly
        dependent local states.
        """
        if np.all(np.array(ijk) == 0):
            s1 = s0
        else:
            s1 = (slice(-1),)
        return s1

    def discretize(self, X):
        """
        Discretize `X`.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is thes patial discretization.
        Returns:
            Float valued field of local states between 0 and 1.
        """
        self.check(X)
        self._select_axes(X)
        H = np.linspace(self.domain[0], self.domain[1], max(self.n_states) + 1)
        X_ = np.maximum(1 - (abs(X[..., None] - H)) / (H[1] - H[0]), 0)
        return X_[..., list(self.n_states)]



"""Wrapper module for PyFFTW and Numpy's fft.

Provide a consistent interface for `rfftn`, `irfftn`, `fftn` and
`ifftn` between Numpy's fft and PyFFTW. PyFFTW takes extra arguments
that are not used by Numpy's fft module, for example, to call `rfftn`
with PyFFTW,

    rfft(np.ascontiguousarray(data),
         axes=axes,
         threads=threads,
         overwrite_input=True,
         planner_effort='FFTW_ESTIMATE',
         avoid_copy=True)

while using Numpy, it is just

    rfft(data,
         axes=axes,)

Furthermore, this module takes care of the logic for selecting whether
to use PyFFTW. If `PYMKS_USE_FFTW` is set in the environment and set to
a positive value then PyFFTW will be used. If `PYMKS_USE_FFTW` is not
set, then the following section in `setup.cfg` is used,

[pymks]
use-fftw = true

If `PYMKS_USE_FFTW` is set to a negative value then Numpy's fft module
is used. If `PYMKS_USE_FFTW` is unavailable and `use-fftw` is set to
false or is unavailable, then Numpy's fft module is used.

"""
import configparser
import os

import numpy.fft as numpy_fft
import numpy as np
from pkg_resources import resource_string
from ..fmks.func import deprecate

@deprecate
def config_use_pyfftw():
    """Determine the value of `use-fftw` in setup.cfg.

    Returns:
      boolean value based on `use-fftw` in setup.cfg
    """
    try:
        FileNotFoundError
    except NameError:
        FileNotFoundError = IOError
    try:
        config_string = resource_string('pymks', '../setup.cfg').decode('utf-8')
    except FileNotFoundError:
        return False
    parser = configparser.ConfigParser()
    parser.read_string(config_string)
    if parser.has_option('pymks', 'use-fftw'):
        return parser.getboolean('pymks', 'use-fftw')
    else:
        return False

@deprecate
def env_use_pyfftw():
    """Determine the value of the `PYMKS_USE_FFTW` environment variable.

    Returns:
      (defined, value): `defined` is bool depending on whether
        `PYMKS_USE_FFTW` is defined, `value` is the bool value it's set
        to (positive if undetermined)
    """
    from distutils.util import strtobool
    env_var = 'PYMKS_USE_FFTW'
    defined = env_var in os.environ
    if defined:
        env_string = os.environ[env_var]
        try:
            value = strtobool(env_string)
        except ValueError:
            value = True
    else:
        value = False
    return defined, value


@deprecate
def import_pyfftw():
    """Isolate pyfftw import

    Returns:
      the pyfftw.builders module
    """
    import pyfftw.builders as pyfftw_fft
    return pyfftw_fft


@deprecate
def choose_fftmodule():
    """Select either PyFFTW or Numpy's FFT

    Logic is:
      - pyfftw when `PYMKS_USE_FFT` is defined and positive
      - pyfftw when `PYMKS_USE_FFT` is defined and undetermined
      - numpy when `PYMKS_USE_FFT` is defined and negative
      - pyfftw when `PYMKS_USE_FFT` is not defined and config has `use_fftw = true`
      - numpy otherwise

    Returns:
      either `pyfftw.builders` or `numpy.fft` depending on logic
    """
    env_defined, env_value = env_use_pyfftw()
    if env_defined:
        if env_value:
            return import_pyfftw()
        else:
            return numpy_fft
    elif config_use_pyfftw():
        return import_pyfftw()
    else:
        return numpy_fft

FFTMODULE = choose_fftmodule()

@deprecate
def arg_wrap(fft_func, **extra_args):
    """Decorator to add kwargs based on fft suite.

    Args:
      fft_func: the fft function to arg wrap
      extra_args: extra args to add to the PyFFTW call

    Returns:
      the wrapped function
    """

    using_fftw = FFTMODULE.__name__.split('.')[0] == 'pyfftw'
    def wrapper(data, axes=None, **kwargs):
        """Wrapper function for Numpy's fft and PyFFTW

        Args:
          data: data to transform
          fftmodule: either numpy.fft or pyfftw.builders
          threads: the threads argument
          extra_args: extra args to add

        Returns:
          the transformed data
        """
        if using_fftw:
            kwargs.update(dict(planner_effort='FFTW_ESTIMATE',
                               avoid_copy=False))
            kwargs.update(extra_args)
            return fft_func(np.ascontiguousarray(data), axes=axes, **kwargs)()
        else:
            kwargs.pop('threads', None)
            return fft_func(data, axes=axes, **kwargs)
    return wrapper

@deprecate
def arg_wrap_overwrite(fft_func):
    """Add the overwrite_input=True argument to arg_wrap

    Args:
      fft_func: the fft function to arg wrap

    Returns:
      the wrapped function
    """
    return arg_wrap(fft_func, overwrite_input=True)


@arg_wrap_overwrite
def rfftn(data, axes=None, **kwargs):
    """Real Fourier transform wrapper

    Args:
      X: microstructure
      axes: the axes over which to perform the transform

    Returns:
      the transformation
    """
    return FFTMODULE.rfftn(data, axes=axes, **kwargs)

@arg_wrap
def irfftn(data, axes=None, axes_shape=None, **kwargs):
    """Inverse real Fourier transform wrapper

    Args:
      data: data to transform
      axes: the axes over which to perform the transform
      axes_shape: the axes shape

    Returns:
      the transformation
    """
    return FFTMODULE.irfftn(data, axes=axes, s=axes_shape, **kwargs)

@arg_wrap_overwrite
def fftn(data, axes=None, **kwargs):
    """Fourier transform wrapper

    Args:
      data: microstructure
      axes: the axes over which to perform the transform

    Returns:
      the transformation
    """
    return FFTMODULE.fftn(data, axes=axes, **kwargs)

@arg_wrap_overwrite
def ifftn(data, axes=None, **kwargs):
    """Inverse Fourier transform wrapper

    Args:
      data: microstructure
      axes: the axes over which to perform the transform

    Returns:
      the transformation
    """
    return FFTMODULE.ifftn(data, axes=axes, **kwargs)



from .imag_ffts import _ImagFFTBasis
import numpy as np
from ..fmks.func import deprecate


class FourierBasis(_ImagFFTBasis):

    r"""
    Discretize a continuous field into `deg` local states using complex
    exponentials such that,

    .. math::

       \frac{1}{\Delta x} \int_s m(h, x) dx =
       \sum_{- L / 2}^{L / 2} m[l, s] exp(l*h*I)

    and the local state space :math:`H` is mapped into the orthogonal domain

    .. math::

       0 \le  H \le 2 \pi

    The mapping of :math:`H` into the domain is done automatically in PyMKS by
    using the `domain` key work argument.

    >>> n_states = 3
    >>> X = np.array([[0., np.pi / 3],
    ...               [2 * np.pi / 3, 2 * np.pi]])
    >>> X_result = np.array(([[[1, 1, 1],
    ...                        [1, np.exp(np.pi / 3 * 1j),
    ...                         np.exp(- np.pi / 3 * 1j)]],
    ...                       [[1, np.exp(2 *np.pi / 3 * 1j),
    ...                         np.exp(- 2 * np.pi / 3 * 1j)],
    ...                         [1, 1, 1]]]))
    >>> basis = FourierBasis(n_states, [0., 2 * np.pi])
    >>> assert(np.allclose(basis.discretize(X), X_result))

    If the microstructure local state values fall outside of the period of
    the specified domain, the values will be mapped back into the domain.

    >>> n_states = 2
    >>> X = np.array([[0, 1.5]])
    >>> four_basis = FourierBasis(n_states, domain=[0, 1])
    >>> X_result = np.array([[[1, 1],
    ...                       [1, -1]]])
    >>> assert np.allclose(X_result, four_basis.discretize(X))
    """

    @deprecate
    def __init__(self, n_states=5, domain=None):

        r"""
        Instantiate a `FourierBasis`

        Args:
            n_states (int, list): The number of local states, or list of local
                states to be used.
            domain (list, optional): indicate the range of expected values for
                the microstructure, default is [0, 2\pi].
        """
        self.n_states = n_states
        if isinstance(self.n_states, int):
            n_states = ((np.arange(self.n_states + 1) // 2)[1:] *
                        (-1) ** np.arange(1, self.n_states + 1))
        if domain is None:
            domain = [0, 2. * np.pi]
        super(FourierBasis, self).__init__(n_states=n_states, domain=domain)

    def discretize(self, X):
        """
        Discretize `X`.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is the spatial discretization.
        Returns:
            Float valued field of Fourier series coefficients in the following
            order 0, 1, ,-1, 2, -2, with shaped `(n_samples, n_x, ...,
            n_states)`.

        >>> X = np.array([[-1, -1./2, 0.,  1./2, 1]])
        >>> f_basis = FourierBasis(4, [-1, 1.])
        >>> X_result = np.array([[[1, 1, 1, 1],
        ...                       [1, 1j, -1j, -1],
        ...                       [1, -1, -1, 1],
        ...                       [1, -1j, 1j, -1],
        ...                       [1, 1, 1, 1]]])
        >>> assert(np.allclose(X_result, f_basis.discretize(X)))

        """
        self._select_axes(X)
        X_scaled = 2. * np.pi * ((X.astype(float) - self.domain[0]) /
                                 (self.domain[1] - self.domain[0]))
        nones = ([None for i in X.shape])
        X_states = np.zeros(X_scaled.shape + (len(self.n_states),))
        X_states[..., :] = np.array(self.n_states)[nones]
        return np.exp(X_scaled[..., None] * X_states * 1j)



"""MKS Correlation Module
For computing auto and cross corelations under assumption
of  periodic or  non-periodic boundary conditions using discreete fourier
transform.


Note that input microstrucure should be 4 dimensional array.
where X=[n_sample,x,y.n_basis]
"""
import numpy as np
from toolz.curried import pipe
from toolz.curried import map as map_, identity
from sklearn.base import TransformerMixin, BaseEstimator
import dask.array as da
from .func import (
    sequence,
    make_da,
    star,
    dapad,
    curry,
    make_da_return,
    make_da_return2,
    dafftshift,
    dafftn,
    daifftn,
    daconj,
    flatten,
)


@make_da
def cross_correlation(arr1, arr2):
    """Returns the non-normalized cross-correlation

    Args:
      arr1: the input field (n_samples,n_x,n_y)
      arr2: the other input field (n_samples,n_x,n_y)

    Returns:
      an nd-array of same dimension as the input field

    >>> import dask.array as da
    >>> x_data = np.asarray([[[1,1,0],
    ...                       [0,0,1],
    ...                       [1,1,0]]])
    >>> chunks = x_data.shape
    >>> x_data = da.from_array(x_data, chunks=chunks)
    >>> y_data = 1 - x_data
    >>> f_data = cross_correlation(x_data, y_data)
    >>> expected = np.asarray([[[ 2,  3,  2],
    ...                         [ 3, 0,  3],
    ...                         [ 2,  3,  2]]])
    >>> assert np.allclose(f_data, expected)
    >>> da.random.seed(42)
    >>> shape = (10, 5, 5)
    >>> chunks = (2, 5, 5)
    >>> x_data = da.random.random(shape, chunks=chunks)
    >>> y_data = 1 - x_data
    >>> f_data = cross_correlation(x_data, y_data)
    >>> assert x_data.chunks == f_data.chunks
    >>> shape = (10, 5, 5)

    When the two input fields have different chunkings

    >>> x_data = da.random.random(shape, chunks=(2,5,5))
    >>> y_data = da.random.random(shape, chunks=(5,5,5))
    >>> f_data = cross_correlation(x_data, y_data)
    >>> print(f_data.chunks)
    ((2, 2, 1, 1, 2, 2), (5,), (5,))

    """
    faxes = lambda x: tuple(np.arange(x.ndim - 1) + 1)

    return pipe(
        arr1,
        dafftn(axes=faxes(arr1)),
        lambda x: daconj(x) * dafftn(arr2, axes=faxes(arr2)),
        daifftn(axes=faxes(arr1)),
        dafftshift(axes=faxes(arr1)),
        lambda x: x.real,
    )


@curry
def auto_correlation(arr):
    """Returns the non-normalized auto-correlation

    Args:
      arr: the input field (n_samples,n_x,n_y)

    Returns:
      an nd-array of same dimension as the input field

    >>> import dask.array as da
    >>> x_data = np.asarray([[[1, 1, 0],
    ...                       [0, 0, 1],
    ...                       [1, 1, 0]]])
    >>> chunks = x_data.shape
    >>> x_data = da.from_array(x_data, chunks=chunks)
    >>> f_data = auto_correlation(x_data)
    >>> expected = [[[3, 2, 3],
    ...              [2, 5, 2],
    ...              [3, 2, 3]]]
    >>> assert np.allclose(f_data, expected)
    >>> shape = (7, 5, 5)
    >>> chunks = (2, 5, 5)
    >>> da.random.seed(42)
    >>> x_data = da.random.random(shape, chunks=chunks)
    >>> f_data = auto_correlation(x_data)
    >>> assert x_data.chunks == f_data.chunks
    >>> print(f_data.chunks)
    ((2, 2, 2, 1), (5,), (5,))

    """
    return cross_correlation(arr, arr)


@curry
def center_slice(x_data, cutoff):
    """Calculate region of interest around the center voxel upto the
    cutoff length

    Args:
      x_data: the data array (n_samples,n_x,n_y), first index is left unchanged
      cutoff: cutoff size

    Returns:
      reduced size array

    >>> a = np.arange(7).reshape(1, 7)
    >>> print(center_slice(a, 2))
    [[1 2 3 4 5]]

    >>> a = np.arange(49).reshape(1, 7, 7)
    >>> print(center_slice(a, 1).shape)
    (1, 3, 3)

    >>> center_slice(np.arange(5), 1)
    Traceback (most recent call last):
    ...
    RuntimeError: Data should be greater than 1D

    """
    if x_data.ndim <= 1:
        raise RuntimeError("Data should be greater than 1D")

    make_slice = sequence(
        lambda x: x_data.shape[1:][x] // 2, lambda x: slice(x - cutoff, x + cutoff + 1)
    )

    return pipe(
        range(len(x_data.shape) - 1),
        map_(make_slice),
        tuple,
        lambda x: (slice(len(x_data)),) + x,
        lambda x: x_data[x],
    )


@curry
@make_da_return2
def two_point_stats(arr1, arr2, periodic_boundary=True, cutoff=None, mask=None):
    r"""Calculate the 2-points stats for two arrays

    The discretized two point statistics are given by

    .. math::

       f[r \; \vert \; l, l'] = \frac{1}{S} \sum_s m[s, l] m[s + r, l']

    where :math:`f[r \; \vert \; l, l']` is the conditional
    probability of finding the local states :math:`l` and :math:`l` at
    a distance and orientation away from each other defined by the
    vector :math:`r`. `See this paper for more details on the
    notation. <https://doi.org/10.1007/s40192-017-0089-0>`_

    The array ``arr1[i]`` (state :math:`l`) is correlated with
    ``arr2[i]`` (state :math:`l'`) for each sample ``i``. Both arrays
    must have the same number of samples and nominal states (integer
    value) or continuous variables.

    To calculate multiple different correlations for each sample, see
    :func:`~pymks.correlations_multiple`.

    To use ``two_point_stats`` as part of a Scikit-learn pipeline, see
    :class:`~pymks.TwoPointCorrelation`.

    Args:
      arr1: array used to calculate cross-correlations, shape
        ``(n_samples,n_x,n_y)``
      arr2: array used to calculate cross-correlations, shape
        ``(n_samples,n_x,n_y)``
      periodic_boundary: whether to assume a periodic boundary
        (default is ``True``)
      cutoff: the subarray of the 2 point stats to keep
      mask: array specifying confidence in the measurement at a pixel,
        shape ``(n_samples,n_x,n_y)``. In range [0,1].

    Returns:
      the snipped 2-points stats

    If both arrays are Dask arrays then a Dask array is returned.

    >>> out = two_point_stats(
    ...     da.from_array(np.arange(10).reshape(2, 5), chunks=(2, 5)),
    ...     da.from_array(np.arange(10).reshape(2, 5), chunks=(2, 5)),
    ... )
    >>> out.chunks
    ((2,), (5,))
    >>> out.shape
    (2, 5)

    If either of the arrays are Numpy then a Numpy array is returned.

    >>> two_point_stats(
    ...     np.arange(10).reshape(2, 5),
    ...     np.arange(10).reshape(2, 5),
    ... )
    array([[ 3.,  4.,  6.,  4.,  3.],
           [48., 49., 51., 49., 48.]])

    Test masking

    >>> array = da.array([[[1, 0 ,0], [0, 1, 1], [1, 1, 0]]])
    >>> mask = da.array([[[1, 1, 1], [1, 1, 1], [1, 0, 0]]])
    >>> norm_mask = da.array([[[2, 4, 3], [4, 7, 4], [3, 4, 2]]])
    >>> expected = da.array([[[1, 0, 1], [1, 4, 1], [1, 0, 1]]]) / norm_mask
    >>> assert np.allclose(
    ...     two_point_stats(array, array, mask=mask, periodic_boundary=False),
    ...     expected
    ... )

    The mask must be in the range 0 to 1.

    >>> array = da.array([[[1, 0], [0, 1]]])
    >>> mask =  da.array([[[2, 0], [0, 1]]])
    >>> two_point_stats(array, array, mask=mask)
    Traceback (most recent call last):
    ...
    RuntimeError: Mask must be in range [0,1]

    """

    cutoff_ = int((np.min(arr1.shape[1:]) - 1) / 2)
    if cutoff is None:
        cutoff = cutoff_
    cutoff = min(cutoff, cutoff_)

    nonperiodic_padder = sequence(
        dapad(
            pad_width=[(0, 0)] + [(cutoff, cutoff)] * (arr1.ndim - 1),
            mode="constant",
            constant_values=0,
        ),
        lambda x: da.rechunk(x, (x.chunks[0],) + x.shape[1:]),
    )

    padder = identity if periodic_boundary else nonperiodic_padder

    if mask is not None:
        if da.max(mask).compute() > 1.0 or da.min(mask).compute() < 0.0:
            raise RuntimeError("Mask must be in range [0,1]")

        mask_array = lambda arr: arr * mask

        normalize = lambda x: x / auto_correlation(padder(mask))
    else:
        mask_array = identity

        if periodic_boundary:
            # The periodic normalization could always be the
            # auto_correlation of the mask. But for the sake of
            # efficiency, we specify the periodic normalization in the
            # case there is no mask.
            normalize = lambda x: x / arr1[0].size
        else:
            normalize = lambda x: x / auto_correlation(padder(np.ones_like(arr1)))

    return sequence(
        map_(mask_array),
        map_(padder),
        list,
        star(cross_correlation),
        normalize,
        center_slice(cutoff=cutoff),
    )([arr1, arr2])


@make_da_return
def correlations_multiple(data, correlations, periodic_boundary=True, cutoff=None):
    r"""Calculate 2-point stats for a multiple auto/cross correlation

    The discretized two point statistics are given by

    .. math::

       f[r \; \vert \; l, l'] = \frac{1}{S} \sum_s m[s, l] m[s + r, l']

    where :math:`f[r \; \vert \; l, l']` is the conditional
    probability of finding the local states :math:`l` and :math:`l'`
    at a distance and orientation away from each other defined by the
    vector :math:`r`. `See this paper for more details on the
    notation. <https://doi.org/10.1007/s40192-017-0089-0>`_

    The correlations are calulated based on pairs given in
    ``correlations`` for each sample.

    To calculate a single correlation for two arrays, see
    :func:`~pymks.two_point_stats`.

    To use ``correlations_multiple`` as part of a Scikit-learn
    pipeline, see :class:`~pymks.TwoPointCorrelation`.

    Args:
      data: the discretized data with shape ``(n_samples, n_x, n_y, n_state)``
      correlations: the correlation pairs, ``[[i0, j0], [i1, j1], ...]``
      periodic_boundary: whether to assume a periodic boundary (default is true)
      cutoff: the subarray of the 2 point stats to keep

    Returns:
      the 2-points stats array

    If ``data`` is a Numpy array then ``correlations_multiple`` will
    return a Numpy array.

    >>> data = np.arange(18).reshape(1, 3, 3, 2)
    >>> out_np = correlations_multiple(data, [[0, 1], [1, 1]])
    >>> out_np.shape
    (1, 3, 3, 2)
    >>> answer = np.array([[[58, 62, 58], [94, 98, 94], [58, 62, 58]]]) + 2. / 3.
    >>> assert np.allclose(out_np[..., 0], answer)

    However, if ``data`` is a Dask array then a Dask array is
    returned.

    >>> data = da.from_array(data, chunks=(1, 3, 3, 2))
    >>> out = correlations_multiple(data, [[0, 1], [1, 1]])
    >>> out.shape
    (1, 3, 3, 2)
    >>> out.chunks
    ((1,), (3,), (3,), (2,))
    >>> assert np.allclose(out[..., 0], answer)

    """

    return pipe(
        range(data.shape[-1]),
        map_(lambda x: (0, x)),
        lambda x: correlations if correlations else x,
        map_(
            lambda x: two_point_stats(
                data[..., x[0]],
                data[..., x[1]],
                periodic_boundary=periodic_boundary,
                cutoff=cutoff,
            )
        ),
        list,
        lambda x: da.stack(x, axis=-1),
        lambda x: da.rechunk(x, x.chunks[:-1] + (-1,)),
    )


class TwoPointCorrelation(BaseEstimator, TransformerMixin):
    """Calculate the 2-point stats for two arrays as part of Scikit-learn
    pipeline.

    Wraps the :func:`~pymks.correlations_multiple` function. See that
    for more complete documentation.

    """

    def __init__(self, correlations=None, periodic_boundary=True, cutoff=None):
        """

        Args:
          correlations: the correlation pairs
          periodic_boundary: whether to assume a periodic boundary (default is true)
          cutoff: the subarray of the 2 point stats to keep


        """

        self.correlations = correlations
        self.periodic_boundary = periodic_boundary
        self.cutoff = cutoff

    def transform(self, data):
        """Transform the data

        Args:
          data: the data to be transformed

        Returns:
          the 2-point stats array
        """
        return correlations_multiple(
            data,
            self.correlations,
            periodic_boundary=self.periodic_boundary,
            cutoff=self.cutoff,
        )

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self


class FlattenTransformer(BaseEstimator, TransformerMixin):
    """Reshape data ready for a PCA.

    Two point correlation data need to be flatten before performing
    PCA. This class flattens the two point correlation data for use in
    a Sklearn pipeline.

    >>> data = np.arange(50).reshape((2, 5, 5))
    >>> FlattenTransformer().transform(data).shape
    (2, 25)

    """

    @staticmethod
    def transform(x_data):
        """Transform the X data

        Args:
          x_data: the data to be transformed

        """
        return flatten(x_data)

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self



"""PyMKS - the materials knowledge system in Python

See the documenation for details at https://pymks.org
"""

try:
    import pyfftw  # pylint: disable=unused-import; # noqa: F401

    # ensure that pyfftw is always imported before numpy to avoid
    # https://github.com/materialsinnovation/pymks/issues/304
except ImportError:
    pass

import os

from ._version import get_versions

from .fmks.data.cahn_hilliard import solve_cahn_hilliard
from .fmks.plot import plot_microstructures
from .fmks.bases.primitive import PrimitiveTransformer
from .fmks.bases.legendre import LegendreTransformer
from .fmks.localization import LocalizationRegressor
from .fmks.localization import ReshapeTransformer
from .fmks.localization import coeff_to_real
from .fmks.data.delta import generate_delta
from .fmks.data.multiphase import generate_multiphase
from .fmks.correlations import FlattenTransformer
from .fmks.correlations import TwoPointCorrelation
from .fmks.data.checkerboard import generate_checkerboard
from .fmks.pair_correlations import paircorr_from_twopoint
from .fmks import GenericTransformer
from .fmks.correlations import two_point_stats, correlations_multiple
from .fmks.data import solve_fe
from .fmks.graspi import graph_descriptors, GraphDescriptors


# the following will be deprecated
from .mks_localization_model import MKSLocalizationModel
from .bases.primitive import PrimitiveBasis
from .bases.legendre import LegendreBasis
from .mks_structure_analysis import MKSStructureAnalysis
from .mks_homogenization_model import MKSHomogenizationModel

MKSRegressionModel = MKSLocalizationModel
DiscreteIndicatorBasis = PrimitiveBasis
ContinuousIndicatorBasis = PrimitiveBasis
# the above will be deprecated


def test(*args):
    r"""Run all the module tests.

    Equivalent to running ``py.test pymks`` in the base of
    PyMKS. Allows an installed version of PyMKS to be tested.

    Args:
      *args: add arguments to pytest

    To test an installed version of PyMKS use

    .. code-block:: bash

       $ python -c "import pymks; pymks.test()"

    """
    import pytest  # pylint: disable=import-outside-toplevel

    path = os.path.join(os.path.split(__file__)[0], "fmks")
    pytest.main(args=[path, "--doctest-modules", "-r s"] + list(args))


__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "__version__",
    "test",
    "solve_cahn_hilliard",
    "plot_microstructures",
    "PrimitiveTransformer",
    "LocalizationRegressor",
    "ReshapeTransformer",
    "coeff_to_real",
    "MKSLocalizationModel",
    "PrimitiveBasis",
    "LegendreBasis",
    "MKSHomogenizationModel",
    "MKSStructureAnalysis",
    "generate_delta",
    "LegendreTransformer",
    "solve_fe",
    "generate_multiphase",
    "FlattenTransformer",
    "TwoPointCorrelation",
    "generate_checkerboard",
    "paircorr_from_twopoint",
    "GenericTransformer",
    "two_point_stats",
    "correlations_multiple",
    "graph_descriptors",
    "GraphDescriptors",
]



from ..bases.imag_ffts import _ImagFFTBasis
import numpy as np
from ..fmks.func import deprecate

class CahnHilliardSimulation(_ImagFFTBasis):
    r"""
    Solve the `Cahn-Hilliard equation
    <https://en.wikipedia.org/wiki/Cahn-Hilliard_equation>`__ for
    multiple samples in arbitrary dimensions. The concentration varies
    from -1 to 1. The equation is given by

    .. math::

       \dot{\phi} = \nabla^2 \left( \phi^3 -
                                    \phi \right) - \gamma \nabla^4 \phi

    The discretiztion scheme used here is from `Chang and Rutenberg
    <http://dx.doi.org/10.1103/PhysRevE.72.055701>`__. The scheme is a
    semi-implicit discretization in time and is given by

    .. math::

       \phi_{t+\Delta t}
       + \left(1 - a_1\right) \Delta t \nabla^2 \phi_{t+\Delta t}
       + \left(1 - a_2\right) \Delta t \gamma \nabla^4 \phi_{t+\Delta t}
       = \phi_t
       - \Delta t \nabla^2 \left(a_1 \phi_t + a_2
                                 \gamma \nabla^2 \phi_t - \phi_t^3 \right)

    where :math:`a_1=3` and :math:`a_2=0`.

    In 1D.

    >>> np.random.seed(99)
    >>> N = 100
    >>> phi = 0.01 * (2 * np.random.random((2, N)) - 1)
    >>> ch = CahnHilliardSimulation(gamma=4)
    >>> for i in range(10000):
    ...     ch.run(phi)
    ...     phi[:] = ch.response
    >>> assert (max(phi.flat) > 2e-3) and (min(phi.flat) < -2e-3)

    In 2D.

    >>> N = 101
    >>> phi = 0.01 * (2 * np.random.random((2, N, N)) - 1)
    >>> ch = CahnHilliardSimulation(gamma=4.)
    >>> for i in range(100):
    ...     ch.run(phi)
    ...     phi[:] = ch.response
    >>> assert (max(phi.flat) > 0.001) and (min(phi.flat) < -0.001)

    In 3D.

    >>> phi = 0.01 * (2 * np.random.random((2, N, N, N)) - 1)
    >>> ch = CahnHilliardSimulation(gamma=4.)
    >>> for i in range(10):
    ...     ch.run(phi)
    ...     phi[:] = ch.response

    >>> assert (max(phi.flat) > 0.0005) and (min(phi.flat) < -0.0005)

    """
    @deprecate
    def __init__(self, dx=0.25, gamma=1., dt=0.001):
        r"""
        Instanitate a CahnHilliardSimulation

        Args:
            dx (float, optional): grid spacing
            dt (float, optional): time step size
            gamma (float, optional): paramater in CH equation

        """
        self.dx = dx
        self.dt = dt
        self.gamma = gamma
        super(CahnHilliardSimulation, self).__init__()

    def run(self, X):
        r"""
        Return the response field

        Args:
            X (ND array): Array representing the concentration field between -1
                and 1 with shape (n_samples, n_x, ...)

        """
        N = X.shape[1]
        if not np.all(np.array(X.shape[1:]) == N):
            raise RuntimeError("X must represent a square domain")

        L = self.dx * N
        k = np.arange(N)

        if N % 2 == 0:
            N1 = N // 2
            N2 = N1
        else:
            N1 = (N - 1) // 2
            N2 = N1 + 1

        k[N2:] = (k - N1)[:N1]
        k = k * 2 * np.pi / L

        i_ = np.indices(X.shape[1:])
        ksq = np.sum(k[i_] ** 2, axis=0)[None]

        self._axes = np.arange(len(X.shape) - 1) + 1
        FX = self._fftn(X)
        FX3 = self._fftn(X ** 3)

        a1 = 3.
        a2 = 0.
        explicit = ksq * (a1 - self.gamma * a2 * ksq)
        implicit = ksq * ((1 - a1) - self.gamma * (1 - a2) * ksq)
        dt = self.dt

        Fy = (FX * (1 + dt * explicit) - ksq * dt * FX3) / (1 - dt * implicit)
        self.response = self._ifftn(Fy).real



"""Test the Legendre basis.
"""

import numpy as np
import dask.array as da
from pymks.fmks.bases import legendre as leg


def polyval(x_data):
    """
        Evaluate Legendre expansion for given input.
    """
    x_data = 4 * x_data - 1
    polys = np.array((np.ones_like(x_data), x_data, (3.0 * x_data ** 2 - 1.0) / 2.0))
    temp = (2.0 * np.arange(3)[:, None, None] + 1.0) / 2.0 * polys
    return np.rollaxis(temp, 0, 3)


def test_1():
    """
        This test compares output using fmks Legendre
        discretization utility to independantly evaluated
        polynomial expansion results.
    """

    data = da.from_array(np.array([[0.25, 0.1], [0.5, 0.25]]), chunks=(2, 2))
    assert np.allclose(
        leg.discretize(data, n_state=3, min_=0.0, max_=0.5, chunks=1).compute(),
        polyval(data),
    )

    data = da.from_array(np.array([[-1, 1], [0, -1]]), chunks=(2, 2))
    assert np.allclose(
        leg.discretize(data, n_state=3, min_=0.0, max_=0.5, chunks=1).compute(),
        polyval(data),
    )


def test_2():
    """
    This test compares output from Legendre discretization to
    known, hardcoded results.
    """
    np.random.seed(3)
    data = da.from_array(np.random.random((1, 3, 3)), chunks=(1, 3, 3))
    data_ = leg.discretize(data, n_state=2, min_=0.0, max_=1.0).compute()
    f_data = np.fft.fftn(data_, axes=(1, 2))
    f_test = np.array(
        [
            [
                [
                    [4.50000000 + 0.0j, -0.79735949 + 0.0],
                    [0.00000000 + 0.0j, -1.00887157 - 1.48005289j],
                    [0.00000000 + 0.0j, -1.00887157 + 1.48005289j],
                ],
                [
                    [0.00000000 + 0.0j, 0.62300683 - 4.97732233j],
                    [0.00000000 + 0.0j, 1.09318216 + 0.10131035j],
                    [0.00000000 + 0.0j, 0.37713401 + 1.87334545j],
                ],
                [
                    [0.00000000 + 0.0j, 0.62300683 + 4.97732233j],
                    [0.00000000 + 0.0j, 0.37713401 - 1.87334545j],
                    [0.00000000 + 0.0j, 1.09318216 - 0.10131035j],
                ],
            ]
        ]
    )
    assert np.allclose(f_data, f_test)



"""Functions for generating data
"""
from ..func import curry


@curry
def solve_fe(x_data, elastic_modulus, poissons_ratio, macro_strain=1.0, delta_x=1.0):
    r"""Solve the elasticity problem

    Use `Sfepy <https://sfepy.org/doc-devel/index.html>`_ to solve a
    linear strain problem in 2D with a varying microstructure on a
    rectangular grid. The rectangle (cube) is held at the negative
    edge (plane) and displaced by 1 on the positive x edge
    (plane). Periodic boundary conditions are applied to the other
    boundaries.

    The boundary conditions on the rectangle (or cube) are given by

    .. math::

       u(L, y) = L \left(1 + \bar{\varepsilon}_{xx}\right)

    .. math::

       u(0, L) = u(0, 0) = 0

    .. math::

       u(x, 0) = u(x, L)

    where :math:`\bar{\varepsilon}_{xx}` is the ``macro_strain``,
    :math:`u` is the displacement in the :math:`x` direction, and
    :math:`L` is the length of the domain. More details about these
    boundary conditions can be found in `Landi et al
    <http://dx.doi.org/10.1016/j.actamat.2010.01.007>`_.

    See the `elasticity notebook for a full set of equations
    <http://pymks.org/en/stable/rst/notebooks/elasticity.html#Elastostatics-Equations>`_.

    ``x_data`` should have integer values that represent the phase of
    the material. The integer values should correspond to the indices
    for the ``elastic_modulus`` and ``poisson_ratio`` sequences and,
    therefore, ``elastic_modulus`` and ``poisson_ratio`` need to be of
    the same length.

    Args:
      x_data: microstructures with shape, ``(n_samples, n_x, ...)``
      elastic_modulus: the elastic modulus in each phase, ``(e0, e1, ...)``
      poissons_ratio: the poissons ratio for each phase, ``(p0, p1, ...)``
      macro_strain: the macro strain, :math:`\bar{\varepsilon}_{xx}`
      delta_x: the grid spacing

    Returns:
      a dictionary of strain, displacement and stress with stress and
      strain of shape ``(n_samples, n_x, ..., 3)`` and displacement shape
      of ``(n_samples, n_x + 1, ..., 2)``

    >>> import numpy as np
    >>> x_data = np.zeros((1, 11, 11), dtype=int)
    >>> x_data[0, :, 1] = 0

    ``x_data`` has values of 0 and 1 and so ``elastic_modulus`` and
    ``poisson_ratio`` must each have 2 entries for phase 0 and phase
    1.

    >>> strain = solve_fe(
    ...     x_data,
    ...     elastic_modulus=(1.0, 10.0),
    ...     poissons_ratio=(0., 0.),
    ...     macro_strain=1.,
    ...     delta_x=1.
    ... )['strain']

    >>> from pymks import plot_microstructures
    >>> fig = plot_microstructures(strain[0, ..., 0], titles=r'$\varepsilon_{xx}$')
    >>> fig.show()  #doctest: +SKIP

    .. image:: strain.png
       :width: 300

    """
    from .elastic_fe import _solve_fe  # pylint: disable=import-outside-toplevel

    return _solve_fe(
        x_data,
        elastic_modulus,
        poissons_ratio,
        macro_strain=macro_strain,
        delta_x=delta_x,
    )



from .mks_structure_analysis import MKSStructureAnalysis
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import Pipeline
import numpy as np
from .fmks.func import deprecate

class MKSHomogenizationModel(MKSStructureAnalysis):

    """
    The `MKSHomogenizationModel` takes in microstructures and a their
    associated macroscopic property, and created a low dimensional structure
    property linkage. The `MKSHomogenizationModel` model is designed to
    integrate with dimensionality reduction techniques and predictive models.

    Attributes:
        degree: Degree of the polynomial used by
            `property_linker`.
        n_components: Number of components used by `dimension_reducer`.
        dimension_reducer: Instance of a dimensionality reduction class.
        property_linker: Instance of class that maps materials property to the
            microstuctures.
        correlations: spatial correlations to be computed
        basis: instance of a basis class
        reduced_fit_data: Low dimensionality representation of spatial
            correlations used to fit the model.
        reduced_predict_data: Low dimensionality representation of spatial
            correlations predicted by the model.
        periodic_axes: axes that are periodic. (0, 2) would indicate that
            axes x and z are periodic in a 3D microstrucure.
        coef_: Array of values that are the coefficients.
        intercept_: Value that are the intercept

    Below is an example of using MKSHomogenizationModel to predict (or
    classify) the type of microstructure using PCA and Logistic Regression.

    >>> import numpy as np
    >>> n_states = 3
    >>> domain = [-1, 1]

    >>> from pymks.bases import LegendreBasis
    >>> leg_basis = LegendreBasis(n_states=n_states, domain=domain)
    >>> from sklearn.decomposition import PCA
    >>> from sklearn.linear_model import LogisticRegression
    >>> reducer = PCA(n_components=3)
    >>> linker = LogisticRegression()
    >>> model = MKSHomogenizationModel(
    ...     basis=leg_basis, dimension_reducer=reducer, property_linker=linker)
    >>> from pymks.datasets import make_cahn_hilliard
    >>> X0, X1 = make_cahn_hilliard(n_samples=50)
    >>> y0 = np.zeros(X0.shape[0])
    >>> y1 = np.ones(X1.shape[0])

    >>> X = np.concatenate((X0, X1))
    >>> y = np.concatenate((y0, y1))

    >>> model.fit(X, y)

    >>> X0_test, X1_test = make_cahn_hilliard(n_samples=3)
    >>> y0_test = model.predict(X0_test)
    >>> y1_test = model.predict(X1_test)
    >>> assert np.allclose(y0_test, [0, 0, 0])
    >>> assert np.allclose(y1_test, [1, 1, 1])
    """
    @deprecate
    def __init__(self, basis=None, dimension_reducer=None, n_components=None,
                 property_linker=None, degree=1, periodic_axes=None,
                 correlations=None, compute_correlations=True, n_jobs=1,
                 store_correlations=False, mean_center=True):
        """
        Create an instance of a `MKSHomogenizationModel`.

        Args:
            basis (class, optional): an instance of a bases class.
            dimension_reducer (class, optional): an instance of a
                dimensionality reduction class with a fit_transform method. The
                default class is PCA.
            property_linker (class, optional): an instance for a machine
                learning class with fit and predict methods.
            n_components (int, optional): number of components kept by the
                dimension_reducer
            degree (int, optional): degree of the polynomial used by
                property_linker.
            periodic_axes (list, optional): axes that are periodic. (0, 2)
                would indicate that axes x and z are periodic in a 3D
                microstrucure.
            correlations (list, optional): list of spatial correlations to
                compute, default is the autocorrelation with the first local
                state and all of its cross correlations. For example if basis
                has basis.n_states=3, correlation would be [(0, 0), (0, 1),
                (0, 2)]. If n_states=[0, 2, 4], the default correlations are
                [(0, 0), (0, 2), (0, 4)] corresponding to the autocorrelations
                for the 0th local state, and the cross correlations with the 0
                and 2 as well as 0 and 4.
            compute_correlations (boolean, optional): If false spatial
                correlations will not be calculated as part of the fit and
                predict methods. The spatial correlations can be passed as `X`
                to both methods, default is True.
            n_jobs (int, optional): number of parallel jobs to run, only used
                if pyfftw is installed.
            store_correlations (boolean, optional): indicate if spatial
                correlations should be stored
            mean_center (boolean, optional): If true the data will be mean
                centered before dimensionality reduction is computed.
        """

        if property_linker is None:
            property_linker = LinearRegression()
        self._linker = Pipeline([('poly', PolynomialFeatures(degree=degree)),
                                ('connector', property_linker)])
        self.degree = degree
        self.property_linker = property_linker
        if not callable(getattr(self.property_linker, "fit", None)):
            raise RuntimeError(
                "property_linker does not have fit() method.")
        if not callable(getattr(self.property_linker, "predict", None)):
            raise RuntimeError(
                "property_linker does not have predict() method.")
        self.compute_correlations = compute_correlations
        self.reduced_fit_data = None
        self.reduced_predict_data = None
        if self.compute_correlations:
            if basis is None:
                raise RuntimeError(('a basis is need to compute spatial ') +
                                   ('correlations'))
        super(MKSHomogenizationModel,
              self).__init__(store_correlations=store_correlations,
                             dimension_reducer=dimension_reducer,
                             correlations=correlations, n_jobs=n_jobs,
                             n_components=n_components, basis=basis,
                             mean_center=mean_center,
                             periodic_axes=periodic_axes)

    @property
    def n_components(self):
        return self._n_components

    @n_components.setter
    def n_components(self, value):
        """Setter for the number of components using by the dimension_reducer
        """
        self._n_components = value
        self.dimension_reducer.n_components = value

    @property
    def degree(self):
        return self._degree

    @degree.setter
    def degree(self, value):
        """Setter for the polynomial degree for property_linker.
        """
        self._degree = value
        self._linker.set_params(poly__degree=value)

    @property
    def coef_(self):
        return self._linker.named_steps['connector'].coef_

    @coef_.setter
    def coef_(self, coef):
        """Setter for the coefficients for property_linker.
        """
        self._linker.named_steps['connector'].coef_ = coef

    @property
    def intercept_(self):
        return self._linker.named_steps['connector'].intercept_

    @intercept_.setter
    def intercept_(self, intercept):
        """Setter for the intercept for property_linker.
        """
        self._linker.named_steps['connector'].intercept_ = intercept

    @property
    def property_linker(self):
        return self._property_linker

    @property_linker.setter
    def property_linker(self, prop_linker):
        """Setter for the property_linker class.
        """
        self._property_linker = prop_linker
        self._linker.set_params(connector=prop_linker)

    def fit(self, X, y, reduce_labels=None, confidence_index=None, size=None):
        """
        Fits data by calculating 2-point statistics from X, preforming
        dimension reduction using dimension_reducer, and fitting the reduced
        data with the property_linker.

        Args:
            X (ND array): The microstructures or spatial correlations, a
                `(n_samples, n_x, ...)` shaped array where `n_samples` is the
                number of samples and `n_x` is the spatial discretization.
            y (1D array): The material property associated with `X`.
            reducer_labels (1D array, optional): label for X used during the
                fit_transform method for the `dimension_reducer`.
            confidence_index (ND array, optional): array with same shape as X
                used to assign a confidence value for each data point.

        Example

        Let's first start with using the microstructure and effective
        properties.

        >>> import numpy as np
        >>> from sklearn.decomposition import PCA
        >>> from sklearn.linear_model import LinearRegression
        >>> from pymks.bases import PrimitiveBasis
        >>> from pymks.stats import correlate

        >>> reducer = PCA(n_components=2)
        >>> linker = LinearRegression()
        >>> prim_basis = PrimitiveBasis(n_states=2, domain=[0, 1])
        >>> correlations = [(0, 0), (1, 1), (0, 1)]
        >>> model = MKSHomogenizationModel(prim_basis,
        ...                                dimension_reducer=reducer,
        ...                                property_linker=linker,
        ...                                correlations=correlations)
        >>> np.random.seed(99)
        >>> X = np.random.randint(2, size=(3, 15))
        >>> y = np.array([1, 2, 3])
        >>> model.fit(X, y)
        >>> X_stats = correlate(X, prim_basis)
        >>> X_reshaped = X_stats.reshape((X_stats.shape[0], -1))
        >>> X_pca = reducer.fit_transform(X_reshaped - np.mean(X_reshaped,
        ...                               axis=1)[:, None])
        >>> assert np.allclose(model.reduced_fit_data, X_pca)

        Now let's use the same method with spatial correlations instead of
        microtructures.

        >>> from sklearn.decomposition import PCA
        >>> from sklearn.linear_model import LinearRegression
        >>> from pymks.bases import PrimitiveBasis
        >>> from pymks.stats import correlate

        >>> reducer = PCA(n_components=2)
        >>> linker = LinearRegression()
        >>> prim_basis = PrimitiveBasis(n_states=2, domain=[0, 1])
        >>> correlations = [(0, 0), (1, 1), (0, 1)]
        >>> model = MKSHomogenizationModel(dimension_reducer=reducer,
        ...                                property_linker=linker,
        ...                                compute_correlations=False)
        >>> np.random.seed(99)
        >>> X = np.random.randint(2, size=(3, 15))
        >>> y = np.array([1, 2, 3])
        >>> X_stats = correlate(X, prim_basis, correlations=correlations)
        >>> model.fit(X_stats, y)
        >>> X_reshaped = X_stats.reshape((X_stats.shape[0], X_stats[0].size))
        >>> X_pca = reducer.fit_transform(X_reshaped - np.mean(X_reshaped,
        ...                               axis=1)[:, None])
        >>> assert np.allclose(model.reduced_fit_data, X_pca)


        """
        if self.compute_correlations:
            if size is not None:
                X = self.basis._reshape_feature(X, size)
            X = self._compute_stats(X, confidence_index)
        X_reshape = self._reduce_shape(X)
        X_reduced = self._fit_transform(X_reshape, reduce_labels)
        self._linker.fit(X_reduced, y)

    def predict(self, X, confidence_index=None):
        """Predicts macroscopic property for the microstructures `X`.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is the spatial discretization.
            confidence_index (ND array, optional): array with same shape as X
                used to assign a confidence value for each data point.

        Returns:
            The predicted macroscopic property for `X`.

        Example

        >>> import numpy as np
        >>> from sklearn.manifold import LocallyLinearEmbedding
        >>> from sklearn.linear_model import BayesianRidge
        >>> from pymks.bases import PrimitiveBasis
        >>> np.random.seed(1)
        >>> X = np.random.randint(2, size=(50, 100))
        >>> y = np.random.random(50)
        >>> reducer = LocallyLinearEmbedding()
        >>> linker = BayesianRidge()
        >>> prim_basis = PrimitiveBasis(2, domain=[0, 1])
        >>> model = MKSHomogenizationModel(prim_basis, n_components=2,
        ...                                dimension_reducer=reducer,
        ...                                property_linker=linker)
        >>> model.fit(X, y)
        >>> X_test = np.random.randint(2, size=(1, 100))

        Predict with microstructures

        >>> y_pred = model.predict(X_test)

        Predict with spatial correlations

        >>> from pymks.stats import correlate
        >>> model.compute_correlations = False
        >>> X_corr = correlate(X, prim_basis, correlations=[(0, 0)])
        >>> model.fit(X_corr, y)
        >>> X_corr_test = correlate(X_test, prim_basis,
        ...                         correlations=[(0, 0)])
        >>> y_pred_stats = model.predict(X_corr_test)
        >>> assert np.allclose(y_pred_stats, y_pred, atol=1e-3)

        """
        if not hasattr(self._linker.get_params()['connector'], "coef_"):
            raise RuntimeError('fit() method must be run before predict().')
        _size = self._size_axes(self.basis)
        X = self.basis._reshape_feature(X, tuple(_size))
        if self.compute_correlations is True:
            X = self._compute_stats(X, confidence_index)
        X_reduced = self._transform(X)
        self.reduced_predict_data = X_reduced
        return self._linker.predict(X_reduced)

    def score(self, X, y, confidence_index=None):
        """
        The score function for the MKSHomogenizationModel. It formats the
        data and uses the score method from the property_linker.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is the spatial discretization.
            y (1D array): The material property associated with `X`.
            confidence_index (ND array, optional): array with same shape as X
                used to assign a confidence value for each data point.

        Returns:
             Score for MKSHomogenizationModel from the selected
             property_linker.
        """
        if not callable(getattr(self._linker, "score", None)):
            raise RuntimeError(
                "property_linker does not have score() method.")
        _size = self._size_axes(self.basis)
        X = self.basis._reshape_feature(X, _size)
        if self.compute_correlations:
            X = self._compute_stats(X, confidence_index)
        X_reduced = self._transform(X)
        return self._linker.score(X_reduced, y)

    def _size_axes(self, basis):
        """Helper function used to get the correct size of the axes when using
        for both periodic and non-periodic axes.
        """
        _size = self.basis._axes_shape
        if self.periodic_axes is None or len(self.periodic_axes) != len(_size):
            _axes = list(range(len(_size)))
            if self.periodic_axes is not None:
                [_axes.remove(a) for a in self.periodic_axes]
            _size = np.ones(len(_size), dtype=int) * _size
            _size[_axes] = _size[_axes] // 2
        return tuple(_size)



"""Test functions for linear elastic FE
"""

import pytest
import numpy as np
from toolz.curried import pipe, first, get
from pymks.fmks.data import solve_fe


def test_3d():
    """Test FE in 3D
    """

    def setone(arr):
        arr[0, :, (arr.shape[0] - 1) // 2] = 1.0
        return arr

    assert pipe(
        5,
        lambda x: np.zeros((1, x, x, x), dtype=int),
        setone,
        solve_fe(elastic_modulus=(1.0, 10.0), poissons_ratio=(0.0, 0.0)),
        lambda x: np.allclose(
            [np.mean(x["strain"][0, ..., i]) for i in range(6)],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        ),
    )


def test_3d_bcs():
    """Test FE in 3D different macro strain
    """
    np.random.seed(8)

    def testing(size, macro_strain):
        return pipe(
            size,
            lambda x: np.random.randint(2, size=(1, x, x, x)),
            solve_fe(
                elastic_modulus=(10.0, 1.0),
                poissons_ratio=(0.3, 0.3),
                macro_strain=macro_strain,
            ),
            get("displacement"),
            first,
            lambda x: (
                np.allclose(x[-1, ..., 0] - x[0, ..., 0], size * macro_strain),
                np.allclose(x[0, ..., 1], x[-1, ..., 1]),
            ),
            all,
        )

    assert testing(4, 0.1)


def test_issue106():
    """Test for missing phases
    """

    def test(x_data):
        solve_fe(x_data, elastic_modulus=(1, 2, 3), poissons_ratio=(0.3, 0.3, 0.3))

    size = 5
    test(np.zeros((1, size, size, size), dtype=int))
    test(np.ones((1, size, size, size), dtype=int))

    with pytest.raises(RuntimeError) as excinfo:
        x_data = np.ones((1, size, size, size), dtype=int)
        x_data[0, 0, 0] = -1
        test(x_data)

    assert str(excinfo.value) == "X must be between 0 and 2."



import numpy as np


def test_elastic_FE_simulation_3D():
    from pymks.datasets.elastic_FE_simulation import ElasticFESimulation
    nx = 5
    ii = (nx - 1) / 2
    X = np.zeros((1, nx, nx, nx), dtype=int)
    X[0, :, ii] = 1
    model = ElasticFESimulation(elastic_modulus=(1., 10.),
                                poissons_ratio=(0., 0.))
    model.run(X)
    solution = [1., 0., 0., 0., 0., 0.]
    assert np.allclose([np.mean(model.strain[0, ..., i]) for i in range(6)],
                       solution)


def test_elastic_FE_simulation_3D_BCs():
    from pymks.datasets.elastic_FE_simulation import ElasticFESimulation
    np.random.seed(8)
    N = 4
    X = np.random.randint(2, size=(1, N, N, N))
    macro_strain = 0.1
    sim = ElasticFESimulation((10.0, 1.0), (0.3, 0.3), macro_strain=0.1)
    sim.run(X)
    u = sim.displacement[0]
    # Check the left/right offset
    assert np.allclose(u[-1, ..., 0] - u[0, ..., 0], N * macro_strain)
    # Check the left/right y-periodicity
    assert np.allclose(u[0, ..., 1], u[-1, ..., 1])


def get_delta_data(nx, ny):
    from pymks.datasets import make_elastic_FE_strain_delta
    return make_elastic_FE_strain_delta(elastic_modulus=(1, 1.1),
                                        poissons_ratio=(0.3, 0.3),
                                        size=(nx, ny))


def get_random_data(nx, ny):
    from pymks.datasets import make_elastic_FE_strain_random
    np.random.seed(8)
    return make_elastic_FE_strain_random(elastic_modulus=(1., 1.1),
                                         poissons_ratio=(0.3, 0.3),
                                         n_samples=1,
                                         size=(nx, ny))


def roll_zip(*args):
    return list(zip(*tuple(np.rollaxis(x, -1) for x in args)))


def test_cahn_hilliard():
    from pymks.datasets.cahn_hilliard_simulation import CahnHilliardSimulation
    from pymks.datasets import make_cahn_hilliard
    from sklearn import metrics
    from pymks import MKSRegressionModel
    from pymks import ContinuousIndicatorBasis

    mse = metrics.mean_squared_error
    n_samples = 100
    n_spaces = 20
    dt = 1e-3
    np.random.seed(0)
    X, y = make_cahn_hilliard(n_samples=n_samples,
                              size=(n_spaces, n_spaces), dt=dt)
    basis = ContinuousIndicatorBasis(10, [-1, 1])
    model = MKSRegressionModel(basis)
    model.fit(X, y)
    X_test = np.array([np.random.random((n_spaces,
                                         n_spaces)) for i in range(1)])
    CHSim = CahnHilliardSimulation(dt=dt)
    CHSim.run(X_test)
    y_test = CHSim.response
    y_pred = model.predict(X_test)
    assert mse(y_test[0], y_pred[0]) < 0.03

if __name__ == '__main__':
    test_MKS_elastic_delta()



import numpy as np


def test_n_componets_from_reducer():
    from pymks import MKSStructureAnalysis
    from pymks import DiscreteIndicatorBasis
    from sklearn.manifold import LocallyLinearEmbedding
    reducer = LocallyLinearEmbedding(n_components=7)
    dbasis = DiscreteIndicatorBasis(n_states=3, domain=[0, 2])
    model = MKSStructureAnalysis(dimension_reducer=reducer, basis=dbasis)
    assert model.n_components == 7


def test_n_components_with_reducer():
    from pymks import MKSStructureAnalysis
    from pymks import DiscreteIndicatorBasis
    from sklearn.manifold import Isomap
    reducer = Isomap(n_components=7)
    dbasis = DiscreteIndicatorBasis(n_states=3, domain=[0, 2])
    model = MKSStructureAnalysis(dimension_reducer=reducer, basis=dbasis,
                                 n_components=9)
    assert model.n_components == 9


def test_n_components_change():
    from pymks import MKSStructureAnalysis
    from pymks import DiscreteIndicatorBasis
    dbasis = DiscreteIndicatorBasis(n_states=2)
    model = MKSStructureAnalysis(basis=dbasis)
    model.n_components = 27
    assert model.n_components == 27


def test_default_n_components():
    from pymks import MKSStructureAnalysis
    from pymks import DiscreteIndicatorBasis
    dbasis = DiscreteIndicatorBasis(n_states=2)
    model = MKSStructureAnalysis(basis=dbasis)
    assert model.n_components == 5


def test_default_dimension_reducer():
    from sklearn.decomposition import PCA
    from pymks import MKSStructureAnalysis
    from pymks import PrimitiveBasis
    model = MKSStructureAnalysis(basis=PrimitiveBasis())
    assert isinstance(model.dimension_reducer, PCA)


def test_default_correlations():
    from pymks import PrimitiveBasis
    from pymks import MKSStructureAnalysis
    prim_basis = PrimitiveBasis(6)
    model_prim = MKSStructureAnalysis(basis=prim_basis)
    assert model_prim.correlations == [(0, 0), (0, 1), (0, 2),
                                       (0, 3), (0, 4), (0, 5)]


def test_set_correlations():
    from pymks import PrimitiveBasis
    from pymks import MKSStructureAnalysis
    test_correlations = [(0, 0), (0, 2), (0, 4)]
    prim_basis = PrimitiveBasis(6)
    model_prim = MKSStructureAnalysis(basis=prim_basis,
                                      correlations=test_correlations)
    assert model_prim.correlations == test_correlations


def test_reshape_X():
    from pymks import MKSStructureAnalysis
    from pymks import PrimitiveBasis
    anaylzer = MKSStructureAnalysis(basis=PrimitiveBasis())
    X = np.arange(18, dtype='float64').reshape(2, 3, 3)
    X_test = np.concatenate((np.arange(-4, 5)[None], np.arange(-4, 5)[None]))
    assert np.allclose(anaylzer._reduce_shape(X), X_test)


def test_set_components():
    from pymks import MKSStructureAnalysis
    from pymks import PrimitiveBasis
    p_basis = PrimitiveBasis(2)
    model = MKSStructureAnalysis(basis=p_basis)
    X = np.random.randint(2, size=(50, 10, 10))
    model.fit(X)
    components = model.components_
    model.components_ = components * 2
    assert np.allclose(model.components_, components * 2)


def test_store_correlations():
    from pymks import MKSStructureAnalysis
    from pymks import PrimitiveBasis
    from pymks.stats import correlate
    p_basis = PrimitiveBasis(2)
    model = MKSStructureAnalysis(basis=p_basis, store_correlations=True, n_components=2)
    X = np.random.randint(2, size=(2, 4, 4))
    model.fit(X)
    X = correlate(X, p_basis, correlations=[(0, 0), (0, 1)])
    assert np.allclose(X, model.fit_correlations)
    X_0 = np.random.randint(2, size=(2, 4, 4))
    model.transform(X_0)
    X_corr_0 = correlate(X_0, p_basis, correlations=[(0, 0), (0, 1)])
    assert np.allclose(X_corr_0, model.transform_correlations)
    X_1 = np.random.randint(2, size=(2, 4, 4))
    model.transform(X_1)
    X_corr_1 = correlate(X_1, p_basis, correlations=[(0, 0), (0, 1)])
    X_corr_ = np.concatenate((X_corr_0, X_corr_1))
    assert np.allclose(X_corr_, model.transform_correlations)


if __name__ == '__main__':
    test_store_correlations()



"""Generate a set of random multiphase microstructures using a gaussian blur filter.
"""

import numpy as np
import dask.array as da
from toolz.curried import pipe
from toolz.curried import map as fmap
from scipy.ndimage.fourier import fourier_gaussian
from ..func import (
    fftn,
    ifftn,
    fftshift,
    ifftshift,
    zero_pad,
    sequence,
    map_blocks,
    curry,
)


conj = curry(np.conjugate)  # pylint: disable=invalid-name
fabs = curry(np.absolute)  # pylint: disable=invalid-name


@curry
def _imfilter(x_data, f_data):
    """
    to convolve f_data over x_data
    """
    return pipe(f_data, ifftshift, fftn, lambda x: x * fftn(x_data), ifftn).real


@curry
def _gaussian_blur_filter(grain_size, domain_size):
    return pipe(
        grain_size,
        lambda x: fourier_gaussian(np.ones(x), np.ones(len(x))),
        fftshift,
        zero_pad(shape=domain_size, chunks=None),
    ).compute()


@curry
def quantile_ndim(arr, quantiles):
    """Use np.quantile across varying quantiles

    Args:
      arr: array to quantize (at least 2D)
      quantiles: same dimensions as arr

    Returns:
      quantized array

    >>> a = np.array([np.arange(10), np.arange(10)])
    >>> q = np.array(((0.2, 0.8), (0.41, 0.59)))
    >>> print(quantile_ndim(a, q))
    [[2 7]
     [4 5]]

    """
    return pipe(
        zip(arr, quantiles),
        fmap(lambda x: np.quantile(*x, axis=-1, interpolation="nearest").T),
        np.stack,
    )


@curry
def _segmentation_values(x_data, volume_fraction, percent_variance):
    def func(shape):
        return (2 * np.random.random(shape) - 1) * percent_variance

    def calc_quantiles():
        return pipe(
            np.cumsum(volume_fraction)[:-1], lambda x: x + func((len(x_data), len(x)))
        )

    return pipe(
        x_data,
        lambda x: np.reshape(x, (x_data.shape[0], -1)),
        quantile_ndim(quantiles=calc_quantiles()),
        lambda x: np.reshape(
            x, [x_data.shape[0]] + [1] * (x_data.ndim - 1) + [len(volume_fraction) - 1]
        ),
    )


@curry
def np_generate(grain_size, volume_fraction, percent_variance, x_blur):
    """Construct a microstructure given a random array.

    Args:
      shape (tuple): (n_sample, n_x, n_y, n_z)
      grain_size (tuple): size of the grain size in the microstructure
      volume_fraction (tuple): the percent volume fraction for each phase
      percent_variance (float): the percent variance for each value of
        volume_fraction

    Returns:
      random multiphase microstructures with shape of `shape`

    """

    seg_values = _segmentation_values(
        volume_fraction=volume_fraction, percent_variance=percent_variance
    )

    return sequence(
        _imfilter(f_data=_gaussian_blur_filter(grain_size, x_blur.shape[1:])),
        lambda x: x[..., None] > seg_values(x),
        lambda x: np.sum(x, axis=-1),
    )(x_blur)


@curry
def generate_multiphase(
    shape, grain_size, volume_fraction, chunks=-1, percent_variance=0.0, seed=None
):  # pylint: disable=too-many-arguments
    """Constructs microstructures for an arbitrary number of phases
    given the size of the domain, and relative grain size.

    Args:
      shape (tuple): shape of the domain ``(n_sample, n_x, n_y)``
      grain_size (tuple): typical expected grain size ``(n_x, n_y)``
      volume_fraction (tuple): the percent volume fraction for each
        phase, which must sum to 1
      chunks (int): chunks_size of the sample index
      percent_variance (float): the percent variance for each value of
        volume_fraction
      seed (int): set the seed value, default is no seed

    Returns:
      A dask array of random-multiphase microstructures
      microstructures for the system of shape given by `shape`.

    Example:

    >>> x_expected = np.array([[[0, 0, 0],
    ...                         [0, 1, 0],
    ...                         [1, 1, 1]]])

    >>> x_actual = generate_multiphase(
    ...     shape=(1, 3, 3),
    ...     grain_size=(1, 1),
    ...     volume_fraction=(0.5, 0.5),
    ...     seed=10
    ... )
    >>> print(x_actual.shape)
    (1, 3, 3)

    >>> assert np.allclose(x_actual, x_expected)

    If `chunks` is not set a Numpy array is returned.

    >>> type(x_actual)
    <class 'numpy.ndarray'>

    If `chunks` is defined a Dask array is returned.

    >>> x = generate_multiphase(
    ...     shape=(2, 3, 3),
    ...     grain_size=(1, 1),
    ...     volume_fraction=(0.5, 0.5),
    ...     chunks=1
    ... )

    >>> print(x.chunks)
    ((1, 1), (3,), (3,))

    """

    if seed is not None:
        da.random.seed(seed)
        np.random.seed(seed)

    if len(grain_size) + 1 != len(shape):
        raise RuntimeError("`shape` should be of length `len(grain_size) + 1`")

    if not np.allclose(np.sum(volume_fraction), 1):
        raise RuntimeError("The terms in the volume fraction list should sum to 1")

    return pipe(
        map_blocks(
            np_generate(grain_size, volume_fraction, percent_variance),
            da.random.random(shape, chunks=(chunks,) + shape[1:]),
            dtype=np.int64,
        ),
        lambda x: x if chunks > 0 else x.compute(),
    )



import numpy as np


def test_n_componets_from_reducer():
    from pymks import MKSHomogenizationModel, DiscreteIndicatorBasis
    from sklearn.manifold import LocallyLinearEmbedding
    reducer = LocallyLinearEmbedding(n_components=7)
    dbasis = DiscreteIndicatorBasis(n_states=3, domain=[0, 2])
    model = MKSHomogenizationModel(dimension_reducer=reducer, basis=dbasis)
    assert model.n_components == 7


def test_n_components_with_reducer():
    from pymks import MKSHomogenizationModel, DiscreteIndicatorBasis
    from sklearn.manifold import Isomap
    reducer = Isomap(n_components=7)
    dbasis = DiscreteIndicatorBasis(n_states=3, domain=[0, 2])
    model = MKSHomogenizationModel(dimension_reducer=reducer, basis=dbasis,
                                   n_components=9)
    assert model.n_components == 9


def test_stress():
    from pymks.datasets import make_elastic_stress_random
    from pymks import MKSHomogenizationModel, DiscreteIndicatorBasis
    sample_size = 200
    grain_size = [(5, 5), (6, 4), (4, 6), (2, 2)]
    n_samples = [sample_size] * len(grain_size)
    elastic_modulus = (410, 200)
    poissons_ratio = (0.28, 0.3)
    macro_strain = 0.001
    size = (21, 21)
    X, y = make_elastic_stress_random(n_samples=n_samples, size=size,
                                      grain_size=grain_size,
                                      elastic_modulus=elastic_modulus,
                                      poissons_ratio=poissons_ratio,
                                      macro_strain=macro_strain, seed=0)
    dbasis = DiscreteIndicatorBasis(n_states=2, domain=[0, 1])
    model = MKSHomogenizationModel(basis=dbasis, n_components=3, degree=3)
    model.fit(X, y)
    test_sample_size = 1
    n_samples = [test_sample_size] * len(grain_size)
    X_new, y_new = make_elastic_stress_random(
        n_samples=n_samples, size=size, grain_size=grain_size,
        elastic_modulus=elastic_modulus, poissons_ratio=poissons_ratio,
        macro_strain=macro_strain, seed=3)
    y_result = model.predict(X_new)
    assert np.allclose(np.round(y_new, decimals=2),
                       np.round(y_result, decimals=2))


def test_n_components_change():
    from pymks import MKSHomogenizationModel, DiscreteIndicatorBasis
    dbasis = DiscreteIndicatorBasis(n_states=2)
    model = MKSHomogenizationModel(basis=dbasis)
    model.n_components = 27
    assert model.n_components == 27


def test_degree_change():
    from pymks import MKSHomogenizationModel, DiscreteIndicatorBasis
    dbasis = DiscreteIndicatorBasis(n_states=2)
    model = MKSHomogenizationModel(basis=dbasis)
    model.degree = 4
    assert model.degree == 4


def test_default_degree():
    from pymks import MKSHomogenizationModel, DiscreteIndicatorBasis
    dbasis = DiscreteIndicatorBasis(n_states=2)
    model = MKSHomogenizationModel(basis=dbasis)
    assert model.degree == 1


def test_default_n_components():
    from pymks import MKSHomogenizationModel, DiscreteIndicatorBasis
    dbasis = DiscreteIndicatorBasis(n_states=2)
    model = MKSHomogenizationModel(basis=dbasis)
    assert model.n_components == 5


def test_default_property_linker():
    from sklearn.linear_model import LinearRegression
    from pymks import MKSHomogenizationModel, PrimitiveBasis
    prim_basis = PrimitiveBasis(n_states=2)
    model = MKSHomogenizationModel(basis=prim_basis)
    assert isinstance(model.property_linker, LinearRegression)


def test_default_dimension_reducer():
    from sklearn.decomposition import PCA
    from pymks import MKSHomogenizationModel
    model = MKSHomogenizationModel(compute_correlations=False)
    assert isinstance(model.dimension_reducer, PCA)


def test_default_correlations():
    from pymks import PrimitiveBasis
    from pymks import MKSHomogenizationModel
    prim_basis = PrimitiveBasis(6)
    model_prim = MKSHomogenizationModel(basis=prim_basis)
    assert model_prim.correlations == [(0, 0), (0, 1), (0, 2),
                                       (0, 3), (0, 4), (0, 5)]


def test_set_correlations():
    from pymks import PrimitiveBasis
    from pymks import MKSHomogenizationModel
    test_correlations = [(0, 0), (0, 2), (0, 4)]
    prim_basis = PrimitiveBasis(6)
    model_prim = MKSHomogenizationModel(basis=prim_basis,
                                        correlations=test_correlations)
    assert model_prim.correlations == test_correlations


def test_coef_setter():
    from pymks import MKSHomogenizationModel
    from pymks import PrimitiveBasis
    p_basis = PrimitiveBasis(2)
    model = MKSHomogenizationModel(basis=p_basis)
    X = np.random.randint(2, size=(50, 10, 10))
    y = np.random.randint(2, size=(50,))
    model.fit(X, y)
    coefs = model.coef_
    model.coef_ = coefs * 2
    assert np.allclose(model.coef_, coefs * 2)


def test_intercept_setter():
    from pymks import MKSHomogenizationModel
    from pymks import PrimitiveBasis
    p_basis = PrimitiveBasis(2)
    model = MKSHomogenizationModel(basis=p_basis)
    X = np.random.randint(2, size=(50, 10, 10))
    y = np.random.randint(2, size=(50,))
    model.fit(X, y)
    intercept = model.intercept_
    model.intercept_ = intercept * 2
    assert np.allclose(model.intercept_, intercept * 2)


if __name__ == '__main__':
    test_stress()
